/******************************************************************************
 * (c) Copyright 2002-2006, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.nkf;

import java.util.List;
import java.util.Set;
import org.netkernel.layer0.urii.ParsedIdentifierImpl;
import org.netkernel.request.IRequest;

/**
 * A Read/Only Wrapper around the calling request that initiated the invocation of the accessor
 * @author  tab
 */
public interface INKFRequestReadOnly
{   
	/** a request to get the data for a URI */
    public static final int VERB_SOURCE=IRequest.VERB_SOURCE;
	/** a request to set the data for a URI */
    public static final int VERB_SINK=IRequest.VERB_SINK;
	/** a request to see if a resource exists */
    public static final int VERB_EXISTS=IRequest.VERB_EXISTS;
	/** a request to delete a resource */
    public static final int VERB_DELETE=IRequest.VERB_DELETE;
	/** a request to create a new resource */
    public static final int VERB_NEW=IRequest.VERB_NEW;
	/** a request to transpresent resource */
    public static final int VERB_TRANSREPRESENT=IRequest.VERB_TRANSREPT;
	/** a request to fragment a resource */
    public static final int VERB_FRAGMENT=IRequest.VERB_FRAGMENT;
	/** a request to resolve */
    public static final int VERB_RESOLVE=IRequest.VERB_RESOLVE;
	/** a request to resolve */
    public static final int VERB_META=IRequest.VERB_META;
    
    
    /** Priorities **/
    public static final int PRIORITY_MIN = IRequest.PRIORITY_MIN;
    public static final int PRIORITY_STD = IRequest.PRIORITY_STD;
    public static final int PRIORITY_MAX = IRequest.PRIORITY_MAX;
    
    /** active identifier arguments **/
    public static final String ARG_ACTIVE_TYPE=ParsedIdentifierImpl.ARG_ACTIVE_TYPE;
    public static final String ARG_ACTIVE_SCHEME=ParsedIdentifierImpl.ARG_ACTIVE_SCHEME;
    
	/** Return the request type (one the static final int constants) */
	int getVerb();
	
	/** Return the Identifier of the request */
	String getIdentifier();
	
	/** Return the representation class that the client expects in the response */
	@SuppressWarnings("unchecked")
	Class getRepresentationClass();
	
	/** Return the number of arguments on the request */
	int getArgumentCount();
	
	/** Get the name of the argument at the given index */
	String getArgumentName(int aIndex);
	
	/** Get the value of the argument at the given index */
	String getArgumentValue(int aIndex);
	
	/** Return true if the given argument is available on the request
	 * @param aName
	 */
	boolean argumentExists(String aName);
	
	/** Return the value of the first instance of given argument, null if it doesn't exist.
	 * @param aName
	 */
	String getArgumentValue(String aName);
	
	/** Returns the primary representation on the request or null if non is supplied
	 */
	Object getPrimary();
	
	/** Returns the primary response on the request or null if non is supplied
	 */
	@SuppressWarnings("unchecked")
	INKFResponseReadOnly getPrimaryAsResponse();
	
	/** Return the Id of the logical endpoint (element) which was resolved, if any, when
	 * the request was resolved to this endpoint
	 * @return String Id or null if no logical endpoint Id was specified.
	 */
	String getResolvedElementId();
	
	/** Return the value for a specific header
	 * @param aKey the key of the header to get the value of
	 * @return
	 */
	Object getHeaderValue(String aKey);
	
	/** Return all the values for a given header. (Useful if there are a variable number)
	 * @param aKey the key of the header to get the value of
	 * @return
	 */
	List<Object> getHeaderValues(String aKey);
	
	/** Return the set of all header keys on this request
	 * @return
	 */
	Set<String> getHeaderKeys();
	
	/** Create an INKFRequest that is a clone of this (INKFRequestReadOnly) request. This request can
	 * then be issued, it is useful for accessors that relay a request on unchanged.
	 * @return
	 * @throws NKFException
	 */
	INKFRequest getIssuableClone() throws NKFException;
		
}