/******************************************************************************
 * (c) Copyright 2002-2006, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.nkf.impl;

import org.netkernel.container.IKernel;
import org.netkernel.container.IMessages;
import org.netkernel.layer0.logging.LogManager;
import org.netkernel.layer0.logging.IApplicationLog;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IPrototypeParameterMeta;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.CheapMap;
import org.netkernel.layer0.util.Utils;
import org.netkernel.request.*;
import org.netkernel.request.impl.ExpiryFactory;
import org.netkernel.request.impl.MetaImpl;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.RequestResponseFieldsWithStickyImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.request.impl.ResponseImpl;
import org.netkernel.scheduler.IInternalResponseMeta;
import org.netkernel.urii.*;

import java.net.URI;
import java.util.*;

/**
 *	The main context implementation implements INKFContext
 * @author  tab
 */
public class NKFContextImpl implements INKFKernelContext, INKFBasicContext, INKFRequestContext, INKFLocale
{
	protected IKernel mKernel;
	private ISpace mSpace;
	private NKFRequestReadOnlyImpl mRequest;
	protected NKFResponseImpl mResponse=RESPONSE_UNSET;
	private Object mEndpoint;
	private IIdentifier mCWU;
	private String mRuntimeSource;
	private List<IResponse> mAdditionalDependencies=null;
	private INKFParametersControl mParameters;
	private IRequestScopeLevel mSubRequestScope;
	private CheapMap mEndpointMetaCache=null;
	private IRequestResponseFields mStickyRequestHeaders=null;
	
	private static final NKFResponseImpl RESPONSE_UNSET = new NKFResponseImpl(null,null,null);
	private static final String META_SCHEME="meta:";
	
	public static void cleanup()
	{	sResourceBundles.clear();
	}

	/** Creates a new instance of NKFContextImpl */
	public NKFContextImpl(IKernel aKernel, ISpace aSpace, NKFRequestReadOnlyImpl aRequest, Object aEndpoint, INKFParametersControl aParameters)
	{	mKernel = aKernel;
		mSpace = aSpace;
		mEndpoint = aEndpoint;
		mRequest=aRequest;
		if (aParameters!=null)
		{	mParameters = aParameters;
		}
		else
		{	mParameters = ParametersImpl.EMPTY;
		}
		
		if (mRequest!=null)
		{	mSubRequestScope=mRequest.getKernelRequest().getRequestScope();
		}
		else
		{	mSubRequestScope=createRootRequestScope();
		}					
	}
	
	IRequestResponseFields getStickyRequestHeaders()
	{	IRequestResponseFields result=mStickyRequestHeaders;
		if (result==null)
		{	IRequest requestIn=getThisKernelRequestOrNull();
			if (requestIn!=null)
			{	result=RequestResponseFieldsWithStickyImpl.getSticky(requestIn.getHeaders());
			}
			else
			{	result=RequestResponseFieldsWithStickyImpl.EMPTY;
			}
			mStickyRequestHeaders=result;
		}
		return result;
	}
	
	
	public void setEndpoint(Object aEndpoint)
	{	mEndpoint=aEndpoint;
		if (mEndpoint instanceof NKFEndpointImpl)
		{	INKFParametersControl params=((NKFEndpointImpl)mEndpoint).getParameters();
			setParameters(params);
		}
	}
	
	public void setParameters(INKFParametersControl aParameters)
	{	mParameters=aParameters;
	}
	
	public ISpace getSpace()
	{	return mSpace;
	}
	
	public void setRequestScope(IRequestScopeLevel aScope)
	{	mSubRequestScope=aScope;
	}
	
	public IRequestScopeLevel getRequestScope()
	{	return mSubRequestScope;
	}

	public IResponse getKernelResponse() throws NKFException
	{	IResponse result=null;
		if (mResponse==RESPONSE_UNSET)
		{	// if no response is explicitly set then response is expired
			IResponseMeta meta = MetaImpl.EXPIRED;
			result = new ResponseImpl(mRequest.getKernelRequest(),null,meta);
		}
		if (result==null && mResponse!=null)
		{	result = mResponse.getKernelResponse();
		}
		return result;
	}
	
	public NKFResponseImpl getResponse()
	{	NKFResponseImpl result=null;
		if (mResponse!=null && mResponse!=RESPONSE_UNSET)
		{	result=mResponse;
		}
		return result;
	}
	
	protected Object getEndpoint()
	{	return mEndpoint;
	}
	
	//*************************
	// INKFKernelContext API
	//*************************
	
	public IKernel getKernel()
	{	return mKernel;
	}
	
	public ISpace getOwningSpace()
	{	return mSpace;
	}
	
	public IRequest getThisKernelRequest()
	{	return mRequest.getKernelRequest();
	}
	
	public IRequest getThisKernelRequestOrNull()
	{	IRequest result=null;
		if (mRequest!=null)
		{	result=mRequest.getKernelRequest();
		}
		return result;
	}
	
	public NKFRequestReadOnlyImpl getThisRequestOrNull()
	{	return mRequest;
	}
	
	public void declareDependency(INKFResponseReadOnly aResponse) throws NKFException
	{	
		NKFResponseReadOnlyImpl resp = (NKFResponseReadOnlyImpl)aResponse;
		if (mAdditionalDependencies==null)
		{	mAdditionalDependencies = new ArrayList(4);
		}
		IResponse kResp=resp.getKernelResponse();
		mAdditionalDependencies.add(kResp);
	}
	
	/** Experimental feature to force a response to be expired regardless of how response is set later
	 *  - maybe useful for getParts and such where response isn't every set but dependencies accumulated 
	 */
	public void declareExpiryDependency()
	{	declareDependency(ExpiryFactory.getNeverExpired(),Integer.MAX_VALUE);
	}
	/*
	public void declareExpiryDependency(long aOffset)
	{	final long expiryTime=getKernel().getTimer().getApproximateTime()+aOffset;
		IExpiry exp=new IExpiry()
		{	public boolean isExpired(long aTime)
			{	return aTime>expiryTime;
			}
		};
		declareDependency(exp,Integer.MAX_VALUE);
	}
	*/
	
	public void declareScopeDependency(int aScope)
	{	declareDependency(ExpiryFactory.getNeverExpired(),aScope);
	}
	
	public void declareDependency(IExpiry aExpiry, int aScope)
	{	IExpiry[] expDet=new IExpiry[]{aExpiry};
		IInternalResponseMeta m = IInternalResponseMeta.create(aExpiry, RequestResponseFieldsImpl.EMPTY, aScope, aScope, 0,0, new IInternalResponseMeta[0],0,expDet);
		IResponse response = new ResponseImpl(null,null,m);

		if (mAdditionalDependencies==null)
		{	mAdditionalDependencies = new ArrayList(4);
		}
		mAdditionalDependencies.add(response);
	}
	

	
	List<IResponse> getAdditionalDependencies()
	{	return mAdditionalDependencies;
	}
	
	public INKFResponseReadOnly wrapKernelResponse(IResponse aResponse)
	{	IRequest req2=aResponse.getRequest();
		String identifier=(req2!=null)?req2.getIdentifier().toString():"dummy";
		INKFRequest req = new NKFRequestImpl(this,identifier);
		return new NKFResponseReadOnlyImpl(aResponse,req,this);
	}

	
	//*************************
	// INKFBasicContext API
	//*************************
	public INKFRequestReadOnly getThisRequest() throws IllegalStateException
	{	if (mRequest==null) throw new IllegalStateException(getErrorMessage("EX_NKF_NO_REQUEST"));
		return mRequest;
	}
	
	public INKFResponse createResponseFrom(Object aRepresentation)
	{	INKFResponse result;
		if (aRepresentation instanceof INKFResponseReadOnly)
		{	result = createResponseFrom((INKFResponseReadOnly)aRepresentation);
		}
		else if (mRequest.getVerb()==INKFRequestReadOnly.VERB_TRANSREPRESENT)
		{	IRequestResponseFields userMeta = ((NKFResponseReadOnlyImpl)getThisRequest().getPrimaryAsResponse()).getUserMeta();
			result = new NKFResponseImpl(aRepresentation,userMeta,this);
		}
		else
		{	result = new NKFResponseImpl(aRepresentation,this);
		}
		setResponse(result);
		return result;
	}
	
	public INKFResponse createResponseFrom(INKFResponseReadOnly aResponse)
	{	NKFResponseImpl resp = new NKFResponseImpl(aResponse,this);
		setResponse(resp);
		return resp;
	}
	
	public void setResponse(INKFResponse aResponse)
	{	mResponse = (NKFResponseImpl)aResponse;
	}
	
	public void setNoResponse()
	{	mResponse = null;
	}
	
	public void unsetResponse()
	{	mResponse=RESPONSE_UNSET;
	}

	public boolean isResponseSet()
	{	return mResponse!=RESPONSE_UNSET;
	}
	
	protected IRequestScopeLevel createRootRequestScope()
	{	return RequestScopeLevelImpl.createRootScopeLevel(getSpace());
	}
	
	public INKFRequest createRequest(String aIdentifier) throws NKFException
	{	NKFRequestImpl req;
		try
		{	String identifier=resolveIdentifier(aIdentifier, true);
			if (identifier.startsWith(META_SCHEME))
			{	String aEndpointId=identifier.substring(META_SCHEME.length());
				IEndpointMeta meta=getEndpointMeta(aEndpointId);
				req= (NKFRequestImpl)createRequestToEndpoint(meta);
			}
			else
			{	req = new NKFRequestImpl(this,identifier);
			}
		}
		catch (LiteralParameterException aLPE)
		{	req=new NKFRequestImpl(this,aIdentifier,aLPE);
		}
		return req;
	}
	
	INKFRequest createAbsoluteRequest(String aIdentifier) throws NKFException
	{	INKFRequest req = new NKFRequestImpl(this,aIdentifier);
		return req;
	}
	
	public INKFRequest cloneThisRequest() throws NKFException
	{	INKFRequestReadOnly thisReq=getThisRequest();
		NKFRequestImpl req= (NKFRequestImpl)createRequest(thisReq.getIdentifier());
		req.setVerb(thisReq.getVerb());
		req.setRepresentationClass(thisReq.getRepresentationClass());
		INKFResponseReadOnly primaryResponse = thisReq.getPrimaryAsResponse();
		if (primaryResponse!=null)
		{	req.addPrimaryArgumentFromResponse(primaryResponse);
		}
		for (String key: thisReq.getHeaderKeys())
		{	for (Object value : thisReq.getHeaderValues(key))
			{	req.setHeader(key, value);
			}
		}
		return req;
	}
	

	public INKFRequest createRequestToEndpoint(String aEndpointId) throws NKFException
	{	if (aEndpointId==null) throw createNKFE("EX_NKF_EPREQ_ARG_ID", null, null);
		IEndpointMeta meta=getEndpointMeta(aEndpointId);
		return createRequestToEndpoint(meta);
	}
	
	public INKFRequest createRequestToEndpoint(IEndpointMeta aMeta) throws NKFException
	{	return new NKFEndpointRequestImpl(this,aMeta);
	}
	
	private IEndpointMeta getEndpointMeta(String aEndpointId) throws NKFException
	{	IEndpointMeta result=null;
		if (mEndpointMetaCache==null)
		{	mEndpointMetaCache=new CheapMap(4);
		}
		else
		{	INKFResponseReadOnly metaResponse=(INKFResponseReadOnly)mEndpointMetaCache.get(aEndpointId);
			if (metaResponse!=null && !metaResponse.isExpired())
			{	result=(IEndpointMeta)metaResponse.getRepresentation();
				declareDependency(metaResponse);
			}
		}
		if (result==null)
		{	INKFRequest metaReq=createAbsoluteRequest(aEndpointId);
			metaReq.setVerb(INKFRequestReadOnly.VERB_META);
			metaReq.setRepresentationClass(IMetaRepresentation.class);
			INKFResponseReadOnly metaResponse=issueRequestForResponse(metaReq);
			Object o=metaResponse.getRepresentation();
			if (o==null || !(o instanceof IEndpointMeta) || ((IEndpointMeta)o).getIdentifierGrammar()==null)
			{	throw createNKFE("EX_NKF_EPREQ_BAD_META", "MSG_NKF_EPREQ_BAD_META", aEndpointId);
			}
			mEndpointMetaCache.put(aEndpointId,metaResponse);
			result=(IEndpointMeta)o;
		}
		return result;
	}
	
	public INKFResponseReadOnly issueRequestForResponse(INKFRequest aRequest) throws NKFException
	{
		if (aRequest==null) throw new IllegalArgumentException(getErrorMessage("EX_NKF_SUBREQ_NULL"));
		
		INKFResponseReadOnly response;
		NKFRequestImpl req=(NKFRequestImpl)aRequest;
		LiteralParameterException lpe = req.getParameterRequest();
		if (lpe!=null)
		{	response = getLiteralParameter(req,lpe);
		}
		else
		{	IRequest request=req.getKernelRequest();
			IResponse kResponse=issueKernelRequest(request);
			response = new NKFResponseReadOnlyImpl(kResponse,aRequest,this);
		}
		return response;
	}

	public Object issueRequest(INKFRequest aRequest) throws NKFException
	{	
		if (aRequest==null) throw new IllegalArgumentException(getErrorMessage("EX_NKF_SUBREQ_NULL"));
		
		Object representation;
		NKFRequestImpl req=(NKFRequestImpl)aRequest;
		LiteralParameterException lpe = req.getParameterRequest();
		if (lpe!=null)
		{	representation = getLiteralParameter(req,lpe).getRepresentation();
		}
		else
		{	IRequest request=((NKFRequestImpl)aRequest).getKernelRequest();
			IResponse response=issueKernelRequest(request);
			representation = response.getRepresentation();
		}
		return representation;
	}
	
	public INKFResponseReadOnly issueRootRequestForResponse(INKFRequest aRequest) 	throws NKFException
	{
		if (aRequest==null) throw new IllegalArgumentException(getErrorMessage("EX_NKF_SUBREQ_NULL"));
		
		INKFResponseReadOnly response;
		NKFRequestImpl req=(NKFRequestImpl)aRequest;
		LiteralParameterException lpe = req.getParameterRequest();
		if (lpe!=null)
		{	response = getLiteralParameter(req,lpe);
		}
		else
		{	IRequest request=req.getKernelRequest();
			request=RequestFactory.cloneAsRootRequest(request);
			IResponse kResponse;
			try
			{	kResponse=mKernel.synchronousExternalRequest(request, getEndpoint(), getSpace());
			} catch (INetKernelException e)
			{	throw new NKFException(e);
			}
			response = new NKFResponseReadOnlyImpl(kResponse,aRequest,this);
		}
		return response;
	}
	
	public INKFAsyncRequestHandle issueAsyncRootRequest(INKFRequest aRequest) throws NKFException
	{
		if (aRequest==null) throw new IllegalArgumentException(getErrorMessage("EX_NKF_SUBREQ_NULL"));
		
		NKFRequestImpl req=(NKFRequestImpl)aRequest;
		NKFAsyncRequestHandleImpl handle  = new NKFAsyncRequestHandleImpl(req,this);
		LiteralParameterException lpe = req.getParameterRequest();
		if (lpe!=null)
		{	NKFResponseReadOnlyImpl response = getLiteralParameter(req, lpe);
			IResponse kResponse = new ResponseImpl(req.getKernelRequest(),response.getRepresentation(),response.getMeta());
			handle.onResponse(kResponse);
		}
		else
		{	IRequest kernelReq = req.getKernelRequest();
			kernelReq=RequestFactory.cloneAsRootRequest(kernelReq);
			kernelReq=RequestFactory.cloneWithRequestor(kernelReq,handle);
			mKernel.onAsyncRequest(kernelReq);
		}
		return handle;		
	}

	
	
	private NKFResponseReadOnlyImpl getLiteralParameter(NKFRequestImpl aRequest, LiteralParameterException aLiteralParameterException) throws NKFException
	{	NKFResponseReadOnlyImpl response;
	
		Object representation = getParamValue(aLiteralParameterException.getName());
		Class desiredRepresentationClass = aRequest.getRepresentationClass();
		if (! desiredRepresentationClass.isAssignableFrom(representation.getClass()))
		{	
			response=(NKFResponseReadOnlyImpl)mParameters.getParameterLiteralAs(aLiteralParameterException.getName(), desiredRepresentationClass, this);
		}
		else
		{	IResponseMeta meta = MetaImpl.NEVER_EXPIRED;
			IResponse kResponse = new ResponseImpl(aRequest.getKernelRequest(),aLiteralParameterException.getRepresentation(),meta);
			response = new NKFResponseReadOnlyImpl(kResponse,aRequest,this);
		}
		return response;
	}
	
	private NKFResponseReadOnlyImpl getCachedLiteralParam(Object aRepresentation, Class aToRepresentation)
	{	return null;
	}
	
	public IResponse issueKernelRequest(IRequest aRequest) throws NKFException
	{	checkLive();
		try
		{	return mKernel.synchronousRequest(aRequest);
		} catch (INetKernelException e)
		{	if (e instanceof NKFException)
			{	throw (NKFException)e;
			}
			else
			{	throw new NKFException(e);
			}
		}
	}
	
	public INKFAsyncRequestHandle issueAsyncRequest(INKFRequest aRequest) throws NKFException
	{	checkLive();
		if (aRequest==null) throw new IllegalArgumentException(getErrorMessage("EX_NKF_SUBREQ_NULL"));
		
		NKFRequestImpl req=(NKFRequestImpl)aRequest;
		NKFAsyncRequestHandleImpl handle  = new NKFAsyncRequestHandleImpl(req,this);
		LiteralParameterException lpe = req.getParameterRequest();
		if (lpe!=null)
		{	NKFResponseReadOnlyImpl response = getLiteralParameter(req, lpe);
			IResponse kResponse = new ResponseImpl(req.getKernelRequest(),response.getRepresentation(),response.getMeta());
			handle.onResponse(kResponse);
		}
		else
		{	IRequest kernelReq = req.getKernelRequest();
			kernelReq=RequestFactory.cloneWithRequestor(kernelReq,handle);
			mKernel.onAsyncRequest(kernelReq);
		}
		return handle;
	}
	
	protected void checkLive() throws NKFException
	{
	}
	
	public void setCWU(String aIdentifier)
	{	mCWU = new SimpleIdentifierImpl(aIdentifier);
	}
	
	public String getCWU()
	{	String result;
		if (mCWU!=null)
		{	result = mCWU.toString();
		}
		else
		{	result = null;
		}
		return result;
	}
	
	IIdentifier getCWUIdentifier()
	{	return mCWU;
	}
	
	public String getRuntimeSourceIdentifier()
	{	return mRuntimeSource;
	}
	
	public void setRuntimeSourceIdentifier(String aRuntimeSource)
	{	mRuntimeSource=aRuntimeSource;
	}
	
	//*************************
	// INKFRequestContext API
	//*************************
	
	
	public void sink(String aIdentifier, Object aRepresentation) throws NKFException
	{	INKFRequest req=createRequest(aIdentifier);
		req.setVerb(INKFRequestReadOnly.VERB_SINK);
		req.addPrimaryArgument(aRepresentation);
		issueRequest(req);
	}
	
	public Object source(String aIdentifier) throws NKFException
	{	INKFRequest req=createRequest(aIdentifier);
		return issueRequest(req);
	}
	public Object source(String aIdentifier, Class aRepresentationClass) throws NKFException
	{	INKFRequest req=createRequest(aIdentifier);
		req.setRepresentationClass(aRepresentationClass);
		return issueRequest(req);
	}
	public INKFResponseReadOnly sourceForResponse(String aIdentifier, Class aRepresentationClass) throws NKFException
	{	INKFRequest req=createRequest(aIdentifier);
		req.setRepresentationClass(aRepresentationClass);
		return issueRequestForResponse(req);
	}
	
	public INKFResponseReadOnly sourceForResponse(String aIdentifier) throws NKFException
	{	INKFRequest req=createRequest(aIdentifier);
		return issueRequestForResponse(req);
	}
	
	
	public Object sourcePrimary(Class aRepresentationClass) throws NKFException
	{	return sourcePrimaryForResponse(aRepresentationClass).getRepresentation();
	}

	public INKFResponseReadOnly sourcePrimaryForResponse(Class aRepresentationClass) throws NKFException
	{	
		INKFResponseReadOnly result;
		
		IResponse sourcedResponse=(IResponse)mRequest.getKernelRequest().getValue(IRequest.TYPE_SOURCED_RESPONSE);
		if (sourcedResponse!=null)
		{	if (aRepresentationClass.isAssignableFrom(sourcedResponse.getRepresentation().getClass()))
			{	result = new NKFResponseReadOnlyImpl(sourcedResponse,null,this);
			}
			else
			{	// we have value but transrept is required
				NKFRequestImpl req=null;
				try
				{	//first try resolving in our space
					req=(NKFRequestImpl)createAbsoluteRequest(mRequest.getIdentifier());
					req.setVerb(INKFRequestReadOnly.VERB_TRANSREPRESENT);
					req.addPrimaryArgumentWithMetaData(sourcedResponse.getRepresentation(),sourcedResponse.getMeta());
					req.setRepresentationClass(aRepresentationClass);
					result = issueRequestForResponse(req);
				}
				catch (NKFException e)
				{	String erf=mKernel.getLogger().format("EX_RESOLUTION_FAILURE");
					if (e.getDeepestId().equals(erf))
					{	// if we fail then try resolving in space of caller
						IRequestScopeLevel parentScope=getThisKernelRequest().getParent().getRequestScope();
						req.setRequestScope(parentScope);
						result = issueRequestForResponse(req);
					}
					else
					{	throw e;
					}
				}
			}
		}
		else
		{	throw new NKFException(Utils.createFormattedException("EX_NKF_NO_PRIMARY", mRequest.getKernelRequest().toString(), mKernel.getLogger()));
		}
		return result;
	}		
	
	public	Object transrept(Object aSource, Class aRepresentationClass) throws NKFException
	{	String id=SimpleIdentifierImpl.getUnique("literal:var").toString();
		INKFRequest req=createRequest(id);
		req.setVerb(INKFRequestReadOnly.VERB_TRANSREPRESENT);
		req.addPrimaryArgument(aSource);
		req.setRepresentationClass(aRepresentationClass);
		return issueRequest(req);
	}
	
	public INKFResponseReadOnly transreptForResponse(Object aSource, Class aRepresentationClass) throws NKFException
	{	String id=SimpleIdentifierImpl.getUnique("literal:var").toString();
		INKFRequest req=createRequest(id);
		req.setVerb(INKFRequestReadOnly.VERB_TRANSREPRESENT);
		req.addPrimaryArgument(aSource);
		req.setRepresentationClass(aRepresentationClass);
		return issueRequestForResponse(req);	
	}
	
	public String requestNew(String aURIBase, Object aOptionalSource) throws NKFException
	{	INKFRequest req=createRequest(aURIBase);
		req.setVerb(INKFRequestReadOnly.VERB_NEW);
		if (aOptionalSource!=null)
		{	req.addPrimaryArgument(aOptionalSource);
		}
		req.setRepresentationClass(IIdentifier.class);
		IIdentifier result=(IIdentifier)issueRequest(req);
		return result.toString();
	}

	public boolean exists(String aIdentifier) throws NKFException
	{	boolean result;
			
		
		try
		{	//make this call to trap aIdentifier being a literal parameter
			String dummy=resolveIdentifier(aIdentifier,true);
			INKFRequest req=createRequest(dummy);
			req.setVerb(INKFRequestReadOnly.VERB_EXISTS);
			Object rep=issueRequest(req);
			result = ((Boolean)rep).booleanValue();
		} catch (NKFException e)
		{	// failed so it doesn't
			result = false;
		}
		catch (ClassCastException e)
		{	// failed because request didn't validly respond to an exists request but didn't return exception
			// so resource must exist
			result=true;
		}
		catch (NullPointerException e)
		{	// null response - bad but probably because it doesn't exist
			result=false;
		}
		catch (LiteralParameterException e)
		{	// literal parameter so it must exist
			result=true;
		}
		return result;
	}
	
	public boolean delete(String aURI) throws NKFException
	{	INKFRequest req=createRequest(aURI);
		req.setVerb(INKFRequestReadOnly.VERB_DELETE);
		Object representation=issueRequest(req);
		boolean result;
		if (representation instanceof Boolean)
		{	result=((Boolean)representation).booleanValue();
		}
		else
		{	result=false;
		}
		return result;
	}
	
	
	public IResolution resolveRequest(INKFRequest aRequest) throws NKFException
	{	IRequest toResolve=((NKFRequestImpl)aRequest).getKernelRequest();
		IRequest req = RequestFactory.createResolveRequest(toResolve, null, null, getThisKernelRequestOrNull(), null);
		IResponse response=issueKernelRequest(req);
		return (IResolution)response.getRepresentation(); 
	}

	
	
	public INKFKernelContext getKernelContext()
	{	return this;
	}
	
	public void handleAsyncResponse(IResponse aResponse)
	{	if (aResponse!=null)
		{	IResponse result = new ResponseImpl(getKernelContext().getThisKernelRequest(), aResponse.getRepresentation(),aResponse.getMeta());
			getThisKernelRequest().getRequestor().onResponse(result);
		}
	}
	
	public void handleAsyncException(Exception aException)
	{	IResponse result = Utils.generateExceptionResponse(getKernelContext().getThisKernelRequest(),aException);
		getThisKernelRequest().getRequestor().onResponse(result);		
	}	
	
	String getErrorMessage(String aKey)
	{	return mKernel.getLogger().format(aKey);
	}
	
	NKFException createNKFE(String aIdKey, String aMessageKey, String aMessageParam)
	{	return createNKFE(aIdKey,aMessageKey,aMessageParam,null);
	}
	
	
	NKFException createNKFE(String aIdKey, String aMessageKey, String aMessageParam1, String aMessageParam2)
	{	IMessages messages=getKernel().getLogger();
		String id=messages.format(aIdKey);
		String message;
		if (aMessageKey!=null)
		{	message= messages.format(aMessageKey,aMessageParam1,aMessageParam2);
		}
		else
		{	message=null;
		}
		return new NKFException(id,message);
	}
	
	// message stuff
	private static final Map<ClassLoader,Object> sResourceBundles = new WeakHashMap<ClassLoader,Object>();
	
	private String lookupMessage(String aKey)
	{	
		String result=null;
		synchronized(sResourceBundles)
		{	
			IRequestScopeLevel scope=mSubRequestScope;
			ClassLoader ecl=mEndpoint.getClass().getClassLoader();
			if (ecl!=scope.getSpace().getClassLoader() && ecl instanceof ClassLoaderWithExports)
			{	//endpoint has a different classloader to first scope level -  probably a prototype
				ISpace dummySpace = new DummySpace((ClassLoaderWithExports)ecl);
				scope=RequestScopeLevelImpl.createNonDurableScopeLevel(dummySpace, scope);
			}
			do
			{	ClassLoader cl=scope.getSpace().getClassLoader();
				ResourceBundle bundle=null;
				if (cl!=null)
				{	Object o=sResourceBundles.get(cl);
					if (o instanceof ResourceBundle)
					{	bundle=(ResourceBundle)o;
					}
					else if (o==null)
					{	try
						{	bundle = ResourceBundle.getBundle("etc.messages",Locale.getDefault(),cl);
							sResourceBundles.put(cl, bundle);
						}
						catch (MissingResourceException e)
						{	sResourceBundles.put(cl,Boolean.FALSE);
						}
					}
				}
				if (bundle!=null)
				{	try
					{	result=bundle.getString(aKey);
						break;
					}
					catch (MissingResourceException e) {;}
				}
				
				scope=scope.getParent();
			}
			while (scope!=null);
		}
		
		return result;
	}
	
	public NKFException createFormattedException(String aId, String aMessage, String aLocation, Throwable aCause, Object... aMessageParams)
	{	String id=null;
		if (aId!=null)
		{	id = formatMessage(aId,(Object[])null);
		}
		String message=null;
		if (aMessage!=null)
		{	message=formatMessage(aMessage,aMessageParams);
		}
		return new NKFException(id,message,aLocation,aCause);
	}
	
	public void logFormatted(int aLevel, String aMessageKey, Object... aParams)
	{	logFormatted(LogManager.DEFAULT_LOG, aLevel, aMessageKey, aParams);
	}
	
	public void logFormatted(String aLog, int aLevel, String aMessageKey, Object... aParams)
	{	String message=formatMessage(aMessageKey, aParams);
		logRaw(aLevel,message);
	}
	
	public void logRaw(int aLevel, String aMessage)
	{	logRaw(LogManager.DEFAULT_LOG,aLevel,aMessage);
	}

	public void logRaw(String aLog, int aLevel, String aMessage)
	{	LogManager al = LogManager.getInstance();
		IApplicationLog log=al.getLog(mSubRequestScope,aLog);
		Object caller=mEndpoint;
		if (mRuntimeSource!=null)
		{	int i=mRuntimeSource.lastIndexOf('/');
			if (i>=0)
			{	caller=mRuntimeSource.substring(i+1);
			}
		}
		log.logRaw(aLevel,caller,aMessage);
	}

	public boolean shouldLog(int aLevel)
	{	return shouldLog(LogManager.DEFAULT_LOG,aLevel);
	}
	
	public boolean shouldLog(String aLog, int aLevel)
	{
		LogManager al = LogManager.getInstance();
		IApplicationLog log=al.getLog(mSubRequestScope,aLog);
		return log.shouldLog(aLevel,mEndpoint);
	}
	
	public String formatMessage(String aMessageKey, Object... aParams)
	{	
		String message=lookupMessage(aMessageKey);
		if (message==null)
		{	if (aParams!=null)
			{	message =mKernel.getLogger().format(aMessageKey,aParams);
			}
			else
			{	message =mKernel.getLogger().format(aMessageKey);
			}
		}
		else
		{	if (aParams!=null)
			{	message=mKernel.getLogger().formatRaw(message,aParams);
			}
		}
		return message;
	}
	
	public String formatRaw(String aMessage, Object... aParams)
	{	String message;
		if (aParams!=null)
		{	message =mKernel.getLogger().formatRaw(aMessage,aParams);
		}
		else
		{	message =aMessage;
		}
		return message;
	}

	public IMetaRepresentation meta(String aIdentifier) throws NKFException
	{	INKFRequest req = createAbsoluteRequest(aIdentifier);
		req.setVerb(INKFRequestReadOnly.VERB_META);
		req.setRepresentationClass(IMetaRepresentation.class);
		return (IMetaRepresentation)issueRequest(req);
	}

	public IMetaRepresentation meta(INKFRequest aRequest) throws NKFException
	{	IRequest metaFor=((NKFRequestImpl)aRequest).getKernelRequest();
		INKFRequest req = createRequest(metaFor.getIdentifier().toString());
		req.addPrimaryArgument(metaFor);
		req.setVerb(INKFRequestReadOnly.VERB_META);
		req.setRepresentationClass(IMetaRepresentation.class);
		return (IMetaRepresentation)issueRequest(req);
	}
	
	/** Internal method for absolutizing relative identifiers of identifiers to arguments or parameters
	 * @param aIdentifier
	 * @return
	 * @throws NKFException
	 */
	String resolveIdentifier(String aIdentifier, boolean aRelative) throws NKFException, LiteralParameterException
	{   
		String result=aIdentifier;
		
		//relativise according to URI syntax if necessary
		if (aRelative)
		{ 	String cwus = getCWU();
			if (cwus!=null && aIdentifier.indexOf(':')<0 && aIdentifier.length()>0)
			{	URI uri=URI.create(aIdentifier);
				URI cwu=URI.create(cwus);
				result=cwu.resolve(uri).toString();
			}
		}
				
	    //Manage this: address space
	    if(aIdentifier.startsWith("arg:"))
	    {	String arg=aIdentifier.substring(4);
	    	NKFRequestReadOnlyImpl parentRequest = getThisRequestOrNull();
			if (parentRequest!=null)
			{	int fragIndex;
				if ((fragIndex=arg.indexOf('#'))>=0)
				{	// we have a fragment on the argument
					String fragment=arg.substring(fragIndex);
					arg=arg.substring(0,fragIndex);
					result = parentRequest.getArgumentValue(arg)+fragment;
				}
				else
				{	result = parentRequest.getArgumentValue(arg);
				}
			}
			else
			{	result=null;
			}
			if (result==null)
			{	throw createNKFE("EX_NKF_REQ_NO_ARG","MSG_NKF_REQ_NO_ARG",arg,aIdentifier); 
			}
		}
	    else if (aIdentifier.startsWith("param:"))
	    {	String name = aIdentifier.substring(6);
	    	if (mParameters.paramExists(name))
	    	{	IPrototypeParameterMeta meta=getParamMeta(name);
				if (meta.getRepresentationalClass()==IIdentifier.class)
				{	Object value=getParamValue(name);
					if (value instanceof IIdentifier)
					{	result = value.toString();
					}
					else
					{	throw new LiteralParameterException(name,value);
					}
				}
				else
				{	throw createFormattedException("EX_NKF_PARAM_NOT_IDENTIFIER","MSG_NKF_PARAM_NOT_IDENTIFIER",null,null,name);
				}
			}
			else
			{	throw createFormattedException("EX_NKF_REQ_NO_PARAM","MSG_NKF_REQ_NO_PARAM",null,null,name);
			}
	    }
	    return result;
	}

	public int getParamCount()
	{	return mParameters.getParamCount();
	}

	public String getParamName(int aIndex)
	{	return mParameters.getParamName(aIndex);
	}

	public Object getParamValue(String aName) throws NKFException
	{	return mParameters.getParamValue(aName);
	}

	public Object getParamValue(int aIndex) throws NKFException
	{	return mParameters.getParamValue(aIndex);
	}

	public boolean paramExists(String aName)
	{	return mParameters.paramExists(aName);
	}

	public List getParamValues(String aName) throws NKFException
	{	return mParameters.getParamValues(aName);
	}

	public IPrototypeParameterMeta getParamMeta(String aName) throws NKFException
	{	return mParameters.getParamMeta(aName);
	}
	
	public void declareParameters(IPrototypeParameterMeta[] aParameterDefn) throws NKFException
	{	if (mParameters.isTolerant())
		{	mParameters = mParameters.cloneWithExtendedDefn(	aParameterDefn);
			mParameters.onCommission(mKernel, mSpace);
		}
		else
		{	throw createNKFE("EX_NKF_UNHANDLED","MSG_NKF_PARAM_NOT_TOLERANT",null); 
		}
	}
	
	public String toString()
	{	StringBuffer sb=new StringBuffer(256);
		sb.append(Utils.shortClassName(this.getClass()));
		sb.append("[ ");
		sb.append(mEndpoint.toString());
		if (mRequest!=null)
		{	sb.append(' ');
			sb.append(mRequest.getKernelRequest().toString());
		}
		sb.append(']');
		return sb.toString();
	}
	
	private static class DummySpace extends NKFSpaceImpl
	{
		private ClassLoaderWithExports mClassLoader;
		
		public DummySpace(ClassLoaderWithExports aClassLoader)
		{	mClassLoader=aClassLoader;
		}
		
		public ClassLoaderWithExports getClassLoader()
		{	return mClassLoader;
		}
	}


}