package org.netkernel.layer0.util;

import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.IHDSPredicate;
import org.netkernel.layer0.representation.impl.HDSNodeListImpl;
import java.util.*;

public class HDSUtils
{
	public static IHDSNodeList inOrderTraversal(IHDSNode aNode)
	{	List<IHDSNode> nodes = new ArrayList<IHDSNode>(128);
		recurseInOrder(aNode,nodes);
		return new HDSNodeListImpl(nodes);
	}
	
	private static void recurseInOrder(IHDSNode aNode, List<IHDSNode> aNodes)
	{	aNodes.add(aNode);
		for (IHDSNode n: aNode.getChildren())
		{	recurseInOrder(n,aNodes);
		}
	}
	
	public static IHDSNodeList filteredInOrderTraversal(IHDSNode aNode, IHDSPredicate aFilter)
	{	List<IHDSNode> nodes = new ArrayList<IHDSNode>(128);
		recurseInOrder(aNode,nodes,aFilter);
		return new HDSNodeListImpl(nodes);
	}
	
	private static void recurseInOrder(IHDSNode aNode, List<IHDSNode> aNodes, IHDSPredicate aFilter)
	{	if (aFilter.matches(aNode))
		{	aNodes.add(aNode);
		}
		for (IHDSNode n: aNode.getChildren())
		{	recurseInOrder(n,aNodes,aFilter);
		}
	}
	
	public static IHDSNode inOrderTraversalNext(IHDSNode aNode, IHDSNode aRoot)
	{
		IHDSNode result=null;
		IHDSNode[] children = aNode.getChildren();
		if (children.length>0)
		{	result = children[0];
		}
		if (result==null)
		{   if (aNode!=aRoot)
			{	result = getNextSibling(aNode);
			}
			else
			{	return null;
			}
		}
		if(result==null)
		{	IHDSNode parent=aNode.getParent();
			while(parent!=aRoot && result==null)
			{	result=getNextSibling(parent);
				parent=parent.getParent();
			}
		}
		return result;
	}
	
	public static IHDSNode getNextSibling(IHDSNode aNode)
	{	IHDSNode result=null;
		IHDSNode parent=aNode.getParent();
		if (parent!=null)
		{	IHDSNode[] siblings=parent.getChildren();
			int l=siblings.length;
			int i=0;
			while (i<l)
			{	if (siblings[i]==aNode)
				{	if (i+1<l)
					{	result=siblings[i+1];
						break;
					}
				}
				i++;
			}
		}
		return result;
	}
	
	public static boolean isDescendantOrSelf(IHDSNode ancestor, IHDSNode testnode)
	{	boolean result=false;
		while(testnode!=null)
		{	if(testnode==ancestor)
			{	result=true;
				break;
			}
			testnode=testnode.getParent();	
		}
		return result;
	}

}
